@@ -1,6 +1,10 @@
-from __future__ import unicode_literals
+from pkg_resources import DistributionNotFound, get_distribution
 
-__version__ = "2.7.0"
+try:
+    __version__ = get_distribution(__name__).version
+except DistributionNotFound:
+    # package is not installed
+    pass
 
 
 def register(
@@ -9,7 +13,7 @@ def register(
     manager_name="history",
     records_class=None,
     table_name=None,
-    **records_config
+    **records_config,
 ):
     """
     Create historical model for `model` and attach history manager to `model`.
