@@ -1,5 +1,3 @@
-from __future__ import unicode_literals
-
 from django.db import models
 
 from simple_history import register
@@ -14,6 +12,24 @@ class AbstractExternal(models.Model):
         abstract = True
 
 
+class AbstractExternal2(models.Model):
+    history = HistoricalRecords(inherit=True, custom_model_name=lambda x: f"Audit{x}")
+
+    class Meta:
+        abstract = True
+        app_label = "external"
+
+
+class AbstractExternal3(models.Model):
+    history = HistoricalRecords(
+        inherit=True, app="external", custom_model_name=lambda x: f"Audit{x}"
+    )
+
+    class Meta:
+        abstract = True
+        app_label = "external"
+
+
 class ExternalModel(models.Model):
     name = models.CharField(max_length=100)
     history = HistoricalRecords()
