@@ -1,15 +1,18 @@
-from __future__ import unicode_literals
-
+import datetime
 import uuid
+
 from django.apps import apps
 from django.conf import settings
 from django.db import models
+from django.db.models.deletion import CASCADE
+from django.db.models.fields.related import ForeignKey
 from django.urls import reverse
 
 from simple_history import register
-from simple_history.models import HistoricalRecords
+from simple_history.models import HistoricalRecords, HistoricForeignKey
+
 from .custom_user.models import CustomUser as User
-from .external.models import AbstractExternal
+from .external.models import AbstractExternal, AbstractExternal2, AbstractExternal3
 
 get_model = apps.get_model
 
@@ -24,13 +27,48 @@ class Poll(models.Model):
         return reverse("poll-detail", kwargs={"pk": self.pk})
 
 
+class PollWithNonEditableField(models.Model):
+    question = models.CharField(max_length=200)
+    pub_date = models.DateTimeField("date published")
+    modified = models.DateTimeField(auto_now=True, editable=False)
+
+    history = HistoricalRecords()
+
+
+class PollWithUniqueQuestion(models.Model):
+    question = models.CharField(max_length=200, unique=True)
+    pub_date = models.DateTimeField("date published")
+
+    history = HistoricalRecords()
+
+
 class PollWithExcludeFields(models.Model):
     question = models.CharField(max_length=200)
     pub_date = models.DateTimeField("date published")
+    place = models.TextField(null=True)
 
     history = HistoricalRecords(excluded_fields=["pub_date"])
 
 
+class PollWithExcludedFieldsWithDefaults(models.Model):
+    question = models.CharField(max_length=200)
+    pub_date = models.DateTimeField("date published")
+    expiration_time = models.DateField(default=datetime.date(2030, 12, 12))
+    place = models.TextField(null=True)
+    min_questions = models.PositiveIntegerField(default=1)
+    max_questions = models.PositiveIntegerField()
+
+    history = HistoricalRecords(
+        excluded_fields=[
+            "pub_date",
+            "expiration_time",
+            "place",
+            "min_questions",
+            "max_questions",
+        ]
+    )
+
+
 class PollWithExcludedFKField(models.Model):
     question = models.CharField(max_length=200)
     pub_date = models.DateTimeField("date published")
@@ -39,6 +77,21 @@ class PollWithExcludedFKField(models.Model):
     history = HistoricalRecords(excluded_fields=["place"])
 
 
+class AlternativePollManager(models.Manager):
+    def get_queryset(self):
+        return super().get_queryset().exclude(id=1)
+
+
+class PollWithAlternativeManager(models.Model):
+    some_objects = AlternativePollManager()
+    all_objects = models.Manager()
+
+    question = models.CharField(max_length=200)
+    pub_date = models.DateTimeField("date published")
+
+    history = HistoricalRecords()
+
+
 class IPAddressHistoricalModel(models.Model):
     ip_address = models.GenericIPAddressField()
 
@@ -59,13 +112,13 @@ class PollWithHistoricalIPAddress(models.Model):
 class CustomAttrNameForeignKey(models.ForeignKey):
     def __init__(self, *args, **kwargs):
         self.attr_name = kwargs.pop("attr_name", None)
-        super(CustomAttrNameForeignKey, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def get_attname(self):
-        return self.attr_name or super(CustomAttrNameForeignKey, self).get_attname()
+        return self.attr_name or super().get_attname()
 
     def deconstruct(self):
-        name, path, args, kwargs = super(CustomAttrNameForeignKey, self).deconstruct()
+        name, path, args, kwargs = super().deconstruct()
         if self.attr_name:
             kwargs["attr_name"] = self.attr_name
         return name, path, args, kwargs
@@ -76,6 +129,26 @@ class ModelWithCustomAttrForeignKey(models.Model):
     history = HistoricalRecords()
 
 
+class CustomAttrNameOneToOneField(models.OneToOneField):
+    def __init__(self, *args, **kwargs):
+        self.attr_name = kwargs.pop("attr_name", None)
+        super().__init__(*args, **kwargs)
+
+    def get_attname(self):
+        return self.attr_name or super().get_attname()
+
+    def deconstruct(self):
+        name, path, args, kwargs = super().deconstruct()
+        if self.attr_name:
+            kwargs["attr_name"] = self.attr_name
+        return name, path, args, kwargs
+
+
+class ModelWithCustomAttrOneToOneField(models.Model):
+    poll = CustomAttrNameOneToOneField(Poll, models.CASCADE, attr_name="custom_poll")
+    history = HistoricalRecords(excluded_field_kwargs={"poll": {"attr_name"}})
+
+
 class Temperature(models.Model):
     location = models.CharField(max_length=200)
     temperature = models.IntegerField()
@@ -124,15 +197,13 @@ class Voter(models.Model):
 class HistoricalRecordsVerbose(HistoricalRecords):
     def get_extra_fields(self, model, fields):
         def verbose_str(self):
-            return "%s changed by %s as of %s" % (
+            return "{} changed by {} as of {}".format(
                 self.history_object,
                 self.history_user,
                 self.history_date,
             )
 
-        extra_fields = super(HistoricalRecordsVerbose, self).get_extra_fields(
-            model, fields
-        )
+        extra_fields = super().get_extra_fields(model, fields)
         extra_fields["__str__"] = verbose_str
         return extra_fields
 
@@ -159,7 +230,7 @@ class Person(models.Model):
         if hasattr(self, "skip_history_when_saving"):
             raise RuntimeError("error while saving")
         else:
-            super(Person, self).save(*args, **kwargs)
+            super().save(*args, **kwargs)
 
 
 class FileModel(models.Model):
@@ -168,15 +239,33 @@ class FileModel(models.Model):
     history = HistoricalRecords()
 
 
+# Set SIMPLE_HISTORY_FILEFIELD_TO_CHARFIELD
+setattr(settings, "SIMPLE_HISTORY_FILEFIELD_TO_CHARFIELD", True)
+
+
+class CharFieldFileModel(models.Model):
+    title = models.CharField(max_length=100)
+    file = models.FileField(upload_to="files")
+    history = HistoricalRecords()
+
+
+# Clear SIMPLE_HISTORY_FILEFIELD_TO_CHARFIELD
+delattr(settings, "SIMPLE_HISTORY_FILEFIELD_TO_CHARFIELD")
+
+
 class Document(models.Model):
     changed_by = models.ForeignKey(
         User, on_delete=models.CASCADE, null=True, blank=True
     )
+
     history = HistoricalRecords()
 
     @property
     def _history_user(self):
-        return self.changed_by
+        try:
+            return self.changed_by
+        except User.DoesNotExist:
+            return None
 
 
 class Paper(Document):
@@ -187,6 +276,12 @@ class Paper(Document):
         self.changed_by = value
 
 
+class RankedDocument(Document):
+    rank = models.IntegerField(default=50)
+
+    history = HistoricalRecords()
+
+
 class Profile(User):
     date_of_birth = models.DateField()
 
@@ -202,7 +297,9 @@ class State(models.Model):
 
 class Book(models.Model):
     isbn = models.CharField(max_length=15, primary_key=True)
-    history = HistoricalRecords(verbose_name="dead trees")
+    history = HistoricalRecords(
+        verbose_name="dead trees", verbose_name_plural="dead trees plural"
+    )
 
 
 class HardbackBook(Book):
@@ -219,6 +316,7 @@ class Library(models.Model):
 
     class Meta:
         verbose_name = "quiet please"
+        verbose_name_plural = "quiet please plural"
 
 
 class BaseModel(models.Model):
@@ -288,6 +386,14 @@ class UnicodeVerboseName(models.Model):
         verbose_name = "\u570b"
 
 
+class UnicodeVerboseNamePlural(models.Model):
+    name = models.CharField(max_length=100)
+    history = HistoricalRecords()
+
+    class Meta:
+        verbose_name_plural = "\u570b"
+
+
 class CustomFKError(models.Model):
     fk = models.ForeignKey(SecondLevelInheritedModel, on_delete=models.CASCADE)
     history = HistoricalRecords()
@@ -343,6 +449,17 @@ class City(models.Model):
     history = HistoricalRecords()
 
 
+class Planet(models.Model):
+    star = models.CharField(max_length=30)
+    history = HistoricalRecords()
+
+    def __str__(self):
+        return self.star
+
+    class Meta:
+        verbose_name = "Planet"
+
+
 class Contact(models.Model):
     name = models.CharField(max_length=30)
     email = models.EmailField(max_length=255, unique=True)
@@ -362,6 +479,23 @@ class ModelWithHistoryInDifferentApp(models.Model):
     history = HistoricalRecords(app="external")
 
 
+class ModelWithHistoryInDifferentDb(models.Model):
+    name = models.CharField(max_length=30)
+    history = HistoricalRecords()
+
+
+class ModelWithHistoryUsingBaseModelDb(models.Model):
+    name = models.CharField(max_length=30)
+    history = HistoricalRecords(use_base_model_db=True)
+
+
+class ModelWithFkToModelWithHistoryUsingBaseModelDb(models.Model):
+    fk = models.ForeignKey(
+        ModelWithHistoryUsingBaseModelDb, on_delete=models.CASCADE, null=True
+    )
+    history = HistoricalRecords(use_base_model_db=True)
+
+
 ###############################################################################
 #
 # Inheritance examples
@@ -442,6 +576,15 @@ class InheritTracking4(TrackedAbstractBaseA):
     pass
 
 
+class BasePlace(models.Model):
+    name = models.CharField(max_length=50)
+    history = HistoricalRecords(inherit=True)
+
+
+class InheritedRestaurant(BasePlace):
+    serves_hot_dogs = models.BooleanField(default=False)
+
+
 class BucketMember(models.Model):
     name = models.CharField(max_length=30)
     user = models.OneToOneField(
@@ -512,7 +655,6 @@ class UUIDRegisterModel(models.Model):
 
 register(UUIDRegisterModel, history_id_field=models.UUIDField(default=uuid.uuid4))
 
-
 # Set the SIMPLE_HISTORY_HISTORY_ID_USE_UUID
 setattr(settings, "SIMPLE_HISTORY_HISTORY_ID_USE_UUID", True)
 
@@ -525,7 +667,6 @@ class UUIDDefaultModel(models.Model):
 # Clear the SIMPLE_HISTORY_HISTORY_ID_USE_UUID
 delattr(settings, "SIMPLE_HISTORY_HISTORY_ID_USE_UUID")
 
-
 # Set the SIMPLE_HISTORY_HISTORY_CHANGE_REASON_FIELD
 setattr(settings, "SIMPLE_HISTORY_HISTORY_CHANGE_REASON_USE_TEXT_FIELD", True)
 
@@ -549,14 +690,57 @@ class CharFieldChangeReasonModel(models.Model):
     history = HistoricalRecords()
 
 
-class CustomNameModel(models.Model):
+class CustomManagerNameModel(models.Model):
+    name = models.CharField(max_length=15)
+    log = HistoricalRecords()
+
+
+#
+# Following classes test the "custom_model_name" option
+#
+
+
+class OverrideModelNameAsString(models.Model):
     name = models.CharField(max_length=15, unique=True)
     history = HistoricalRecords(custom_model_name="MyHistoricalCustomNameModel")
 
 
-class CustomManagerNameModel(models.Model):
-    name = models.CharField(max_length=15)
-    log = HistoricalRecords()
+class OverrideModelNameAsCallable(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+    history = HistoricalRecords(custom_model_name=lambda x: f"Audit{x}")
+
+
+class AbstractModelCallable1(models.Model):
+    history = HistoricalRecords(inherit=True, custom_model_name=lambda x: f"Audit{x}")
+
+    class Meta:
+        abstract = True
+
+
+class OverrideModelNameUsingBaseModel1(AbstractModelCallable1):
+    name = models.CharField(max_length=15, unique=True)
+
+
+class OverrideModelNameUsingExternalModel1(AbstractExternal2):
+    name = models.CharField(max_length=15, unique=True)
+
+
+class OverrideModelNameUsingExternalModel2(AbstractExternal3):
+    name = models.CharField(max_length=15, unique=True)
+
+
+class OverrideModelNameRegisterMethod1(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+
+
+register(
+    OverrideModelNameRegisterMethod1,
+    custom_model_name="MyOverrideModelNameRegisterMethod1",
+)
+
+
+class OverrideModelNameRegisterMethod2(models.Model):
+    name = models.CharField(max_length=15, unique=True)
 
 
 class ForeignKeyToSelfModel(models.Model):
@@ -567,3 +751,106 @@ class ForeignKeyToSelfModel(models.Model):
         "self", null=True, related_name="+", on_delete=models.CASCADE
     )
     history = HistoricalRecords()
+
+
+class Street(models.Model):
+    name = models.CharField(max_length=150)
+    log = HistoricalRecords(related_name="history")
+
+
+class ManyToManyModelOther(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+
+
+class BulkCreateManyToManyModel(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+    other = models.ManyToManyField(ManyToManyModelOther)
+    history = HistoricalRecords()
+
+
+class ModelWithExcludedManyToMany(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+    other = models.ManyToManyField(ManyToManyModelOther)
+    history = HistoricalRecords(excluded_fields=["other"])
+
+
+class ModelWithSingleNoDBIndexUnique(models.Model):
+    name = models.CharField(max_length=15, unique=True, db_index=True)
+    name_keeps_index = models.CharField(max_length=15, unique=True, db_index=True)
+    history = HistoricalRecords(no_db_index="name")
+
+
+class ModelWithMultipleNoDBIndex(models.Model):
+    name = models.CharField(max_length=15, db_index=True)
+    name_keeps_index = models.CharField(max_length=15, db_index=True)
+    fk = models.ForeignKey(
+        "Library", on_delete=models.CASCADE, null=True, related_name="+"
+    )
+    fk_keeps_index = models.ForeignKey(
+        "Library", on_delete=models.CASCADE, null=True, related_name="+"
+    )
+    history = HistoricalRecords(no_db_index=["name", "fk", "other"])
+
+
+class TestOrganization(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+
+
+class TestOrganizationWithHistory(models.Model):
+    name = models.CharField(max_length=15, unique=True)
+    history = HistoricalRecords()
+
+
+class TestParticipantToHistoricOrganization(models.Model):
+    """
+    Non-historic table foreign key to historic table.
+
+    In this case it should simply behave like ForeignKey because
+    the origin model (this one) cannot be historic, so foreign key
+    lookups are always "current".
+    """
+
+    name = models.CharField(max_length=15, unique=True)
+    organization = HistoricForeignKey(
+        TestOrganizationWithHistory, on_delete=CASCADE, related_name="participants"
+    )
+
+
+class TestHistoricParticipantToOrganization(models.Model):
+    """
+    Historic table foreign key to non-historic table.
+
+    In this case it should simply behave like ForeignKey because
+    the origin model (this one) can be historic but the target model
+    is not, so foreign key lookups are always "current".
+    """
+
+    name = models.CharField(max_length=15, unique=True)
+    organization = HistoricForeignKey(
+        TestOrganization, on_delete=CASCADE, related_name="participants"
+    )
+    history = HistoricalRecords()
+
+
+class TestHistoricParticipanToHistoricOrganization(models.Model):
+    """
+    Historic table foreign key to historic table.
+
+    In this case as_of queries on the origin model (this one)
+    or on the target model (the other one) will traverse the
+    foreign key relationship honoring the timepoint of the
+    original query.  This only happens when both tables involved
+    are historic.
+
+    NOTE: related_name has to be different than the one used in
+          TestParticipantToHistoricOrganization as they are
+          sharing the same target table.
+    """
+
+    name = models.CharField(max_length=15, unique=True)
+    organization = HistoricForeignKey(
+        TestOrganizationWithHistory,
+        on_delete=CASCADE,
+        related_name="historic_participants",
+    )
+    history = HistoricalRecords()
